#include "map.h"

#include <algorithm>
#include <tuple>

game_map::game_map(RNG *rng, const feature enabled_features):
        enabled_features{enabled_features} {
        map.reserve(MAP_HEIGHT * MAP_WIDTH);

        for (int i = MAP_HEIGHT * MAP_WIDTH; i > 0; --i)
                map.push_back(' ');

        // Note: during generation, walls DO NOT count as being in the rooms
        std::vector<std::pair<position, int>> room_dims = gen_room_dims(rng);

        // width height top left
        std::vector<std::pair<position, position>> room_data = distr_rooms(rng,
                                                room_dims);

        for (size_t r = 0; r < room_data.size(); ++r) {
                for (int x = 0; x < room_data[r].first.x; ++x)
                        for (int y = 0; y < room_data[r].first.y; ++y)
                                map[remap_position({room_data[r].second.x + x,
                                                    room_data[r].second.y + y})] = r;

                for (int x = 0; x < room_data[r].first.x; ++x) {
                        map[remap_position({room_data[r].second.x + x,
                                            room_data[r].second.y - 1})] = '-';
                        map[remap_position({room_data[r].second.x + x,
                                            room_data[r].second.y +
                                            room_data[r].first.y + 1})] = '-';
                }

                for (int y = -1; y <= room_data[r].first.y; ++y) {
                        map[remap_position({room_data[r].second.x - 1,
                                            room_data[r].second.y + y})] = '|';
                        map[remap_position({room_data[r].second.x +
                                            room_data[r].first.x + 1,
                                            room_data[r].second.y + y})] = '|';
                }
        }

        map.shrink_to_fit();
}

std::vector<std::pair<position, position>> game_map::distr_rooms(RNG *rng,
std::vector<std::pair<position, int>> &room_dims) {
        std::vector<std::pair<position, position>> result;
        result.reserve(room_dims.size());

        int max_layer = room_dims[room_dims.size() - 1].second;
        // left, right, height
        std::vector<std::pair<position, int>> layer_data;
        layer_data.reserve(max_layer + 1);

        // distributing rooms horizontally
        for (int layer = 0; layer <= max_layer; ++layer) {
                int l = INF;
                int r = 0;
                int layer_height = 0;

                // get the interval for the current layer
                //  and the max height of the layer
                for (std::size_t i = 0; i < room_dims.size(); ++i)
                        if (room_dims[i].second == layer) {
                                l = std::min(l, (int)i);
                                r = std::max(r, (int)i);
                                layer_height = std::max(layer_height,
                                                        room_dims[i].first.y);
                        }

                layer_data.push_back({{l, r}, layer_height});

                // distribute the current layer
                if (l == r) {
                        result.push_back({{
                                        0, rng->rand_under(ACTUAL_MAP_WIDTH -
                                                           room_dims[l].first.x + 1)
                                },
                                room_dims[l].first});
                        continue;
                }

                int left = 0;
                int right = ACTUAL_MAP_WIDTH;

                // every time, distribute the last one first
                for (int i = l; i < r; ++i)
                        left += MIN_ROOM_SPACING + room_dims[i].first.x;

                for (int i = r; i >= l; --i) {
                        int offset = rng->rand_between(left, right - room_dims[i].first.x + 1);
                        result.push_back({{0, offset}, room_dims[i].first});

                        right = offset - MIN_ROOM_SPACING;

                        if (i != l)
                                left -= MIN_ROOM_SPACING + room_dims[i - 1].first.x;
                }
        }

        // distributing rooms vertically
        int top = 0;
        int bottom = ACTUAL_MAP_HEIGHT;

        for (int i = 0; i < max_layer; ++i)
                top += MIN_ROOM_SPACING + layer_data[i].second;

        for (int i = max_layer; i >= 0; --i) {
                int offset = rng->rand_between(top, bottom - layer_data[i].second + 1);

                for (int j = layer_data[i].first.x; j <= layer_data[i].first.y; ++j)
                        result[j].second.y = offset;

                bottom = offset - MIN_ROOM_SPACING;

                if (i != 0)
                        top -= MIN_ROOM_SPACING + layer_data[i - 1].second;
        }

        // jitter
        //jitter(rng, result);

        //add padding
        for (auto room : result)
                room.first += {MAP_PADDING, MAP_PADDING};

        return result;
}

void game_map::jitter(RNG *rng,
                      std::vector<std::pair<position, position>> &rooms) {
        for (auto target = rooms.rbegin(); target != rooms.rend(); ++target) {
                int t = -INF;
                int b = INF;
                int l = -INF;
                int r = INF;

                for (auto i : rooms)
                        if (i != *target) {
                                if (overlap_x(*target, i)) {
                                        if (target->second.y > i.second.y)
                                                t = std::max(t,
                                                             i.second.y + i.first.y
                                                             + MIN_ROOM_SPACING);
                                        else
                                                b = std::min(b,
                                                             i.second.y
                                                             - MIN_ROOM_SPACING);
                                }

                                if (overlap_y(*target, i)) {
                                        if (target->second.x > i.second.x)
                                                l = std::max(l,
                                                             i.second.x + i.first.x
                                                             + MIN_ROOM_SPACING);
                                        else
                                                r = std::min(r,
                                                             i.second.x
                                                             - MIN_ROOM_SPACING);
                                }
                        }

                target->second.x = rng->rand_between(l, r - target->first.x + 1);
                target->second.y = rng->rand_between(t, b - target->first.y + 1);
        }
}

bool game_map::overlap_x(std::pair<position, position> &room1,
                         std::pair<position, position> &room2) {
        return (room1.second.x >= room2.second.x &&
                room1.second.x - 1 <= room2.second.x + room2.first.x + 1) ||
               (room2.second.x >= room1.second.x &&
                room2.second.x - 1 <= room1.second.x + room1.first.x + 1);
}

bool game_map::overlap_y(std::pair<position, position> &room1,
                         std::pair<position, position> &room2) {
        return (room1.second.y >= room2.second.y &&
                room1.second.y - 1 <= room2.second.y + room2.first.y + 1) ||
               (room2.second.y >= room1.second.y &&
                room2.second.y - 1 <= room1.second.y + room1.first.y + 1);
}

position game_map::random_size(int min_width, int min_height,
                               int max_width, int max_height,
                               RNG *rng) {
        return {rng->rand_between(min_width, max_width + 1),
                rng->rand_between(min_height, max_height + 1)};
}

game_map::game_map(const std::string &map_data,
                   RNG *rng, const feature enabled_features):
        enabled_features{enabled_features} {
        int map_size = MAP_HEIGHT * MAP_WIDTH;
        map.reserve(map_size);

        for (int i = 0; i < map_size; ++i) {
                if (map_data[i] >= '0' && map_data[i] <= '9')
                        map.push_back(map_data[i] - '0');
                else
                        map.push_back(map_data[i]);

                if ((map_data[i] >= '0' && map_data[i] <= '9') ||
                    map_data[i] == '+' || map_data[i] == '#')
                        empty_spots.push_back(remap_index(i));
        }

        down_stairs = rng->get_rand_in_vector(empty_spots);


        if (enabled_features | FEATURE_REVISIT) {
                auto tmp = remove_from_list(empty_spots, down_stairs);
                up_stairs = rng->get_rand_in_vector(tmp);

                map[remap_position(up_stairs)] = '<';
                map[remap_position(down_stairs)] = '>';
        } else {
                map[remap_position(down_stairs)] = '\\';
        }
}

const position_list game_map::get_available_positions() const {
        return empty_spots;
}

bool game_map::is_available(const position &pos) const {
        int idx = remap_position(pos);
        return map[idx] <= 9 || map[idx] == '\\' || map[idx] == '>' ||
               map[idx] == '<' || map[idx] == '+' || map[idx] == '#';
}

void game_map::print(display &out) const {
        for (std::size_t i = 0; i < map.size(); ++i)
                out.print_char(remap_index(i), map[i] <= 9 ? '.' : map[i],
                               map[i] == '\\' || map[i] == '>' ||
                               map[i] == '<' ? COLOR_PAIR(COLOR_BLUE) : A_NORMAL);
}

position game_map::get_up_stairs() const {
        return up_stairs;
}

position game_map::get_down_stairs() const {
        return down_stairs;
}

int game_map::get_up_stairs_room() const {
        return map[remap_position(up_stairs)];
}

int game_map::get_down_stairs_room() const {
        return map[remap_position(down_stairs)];
}

const std::vector<position_list>game_map::get_room_list() const {
        std::vector<position_list> result;
        result.reserve(10);

        for (int i = 0; i < 10; ++i)
                result.push_back({});

        for (std::size_t i = 0; i < map.size(); ++i)
                if (map[i] <= 9)
                        result[map[i]].push_back(remap_index(i));

        result.shrink_to_fit();

        for (std::size_t i = 0; i < result.size(); ++i)
                result[i].shrink_to_fit();

        return result;
}

std::vector<std::pair<position, int>> game_map::gen_room_dims(RNG *rng) {
        position_list room_dim_list;
        room_dim_list.reserve(MAX_ROOM_CNT);
        std::vector<std::pair<position, int>> result;
        result.reserve(MAX_ROOM_CNT);

        for (int i = 0; i < MAX_ROOM_CNT; ++i)
                room_dim_list.push_back(random_size(MIN_ROOM_WIDTH, MIN_ROOM_HEIGHT,
                                                    MAX_ROOM_WIDTH, MAX_ROOM_HEIGHT,
                                                    rng));

        int curr_top  = 0;
        int curr_left = 0;
        // index in result
        int curr_layer_idx = 0;
        int curr_layer_cnt = 0;

        for (int i = 0; i < MAX_ROOM_CNT &&
             curr_top <= ACTUAL_MAP_HEIGHT &&
             curr_left <= ACTUAL_MAP_WIDTH; ++i) {
                // DO NOT update curr_top until curr_left is exhausted
                if (curr_top + MIN_ROOM_SPACING +
                    room_dim_list[i].y + HEIGHT_RESERVED <= ACTUAL_MAP_HEIGHT &&
                    curr_left + MIN_ROOM_SPACING +
                    room_dim_list[i].x + WIDTH_RESERVED <= ACTUAL_MAP_WIDTH) {
                        result.push_back({room_dim_list[i], curr_layer_cnt});
                        curr_left += MIN_ROOM_SPACING + room_dim_list[i].x;
                        continue;
                }

                int tmp = curr_top;

                for (std::size_t j = curr_layer_idx; j < result.size(); ++j)
                        tmp = std::max(curr_top + result[j].first.y + MIN_ROOM_SPACING,
                                       tmp);

                curr_top = tmp;
                curr_layer_idx = result.size();
                curr_left = 0;
                ++curr_layer_cnt;

                if (curr_top + MIN_ROOM_SPACING +
                    room_dim_list[i].y + HEIGHT_RESERVED <= ACTUAL_MAP_HEIGHT &&
                    curr_left + MIN_ROOM_SPACING +
                    room_dim_list[i].x + WIDTH_RESERVED <= ACTUAL_MAP_WIDTH) {
                        result.push_back({room_dim_list[i], curr_layer_cnt});
                        curr_left += MIN_ROOM_SPACING + room_dim_list[i].x;
                }
        }

        result.shrink_to_fit();

        return result;
}

position random_size(int min_width, int min_height,
                     int max_width, int max_height,
                     RNG *rng) {
        return {rng->rand_between(min_width, max_width + 1),
                rng->rand_between(min_height, max_height + 1)};
}
